import numpy as np
import xarray as xr

from ..utils.jaggedarray import flatten_jagged_array


MESH_ATTRS = {
    'cf_role': 'mesh_topology',
    'long_name': 'Topology data of 2D unstructured mesh',
    'topology_dimension': 2,
    'node_coordinates': 'x_of_node y_of_node',
    'face_node_connectivity': 'nodes_at_patch',
    'face_dimension': 'patch',
    'face_edge_connectivity': 'links_at_patch',
    'edge_node_connectivity': 'nodes_at_link',
    'edge_dimension': 'link',
}

DUAL_MESH_ATTRS = {
    'cf_role': 'mesh_topology',
    'long_name': 'Topology data of 2D unstructured mesh',
    'topology_dimension': 2,
    'node_coordinates': 'x_of_corner y_of_corner',
    'face_node_connectivity': 'corners_at_cell',
    'face_dimension': 'cell',
    'face_edge_connectivity': 'faces_at_cell',
    'edge_node_connectivity': 'corners_at_face',
    'edge_dimension': 'face',
}


def ugrid_from_structured_quad(coords, shape=None):
    from .structured_quad.structured_quad import (setup_nodes_at_link,
                                                  setup_links_at_patch,
                                                  setup_node_coords_structured)
    node_y_and_x = setup_node_coords_structured(coords, shape=shape)
    shape = node_y_and_x[0].shape

    return ugrid_from_unstructured(
        node_y_and_x,
        setup_nodes_at_link(shape),
        setup_links_at_patch(shape)) 


def ugrid_from_rectilinear(coords):
    from .structured_quad.structured_quad import setup_node_coords_rectilinear
    node_y_and_x = setup_node_coords_rectilinear(coords)
    return ugrid_from_structured_quad(node_y_and_x)


def ugrid_from_uniform_rectilinear(shape, spacing=1., origin=0.):
    from .structured_quad.structured_quad import setup_node_coords
    node_y_and_x = setup_node_coords(shape, spacing, origin)
    return ugrid_from_structured_quad(node_y_and_x)


def load_ugrid(ugrid):
    if not isinstance(ugrid, xr.Dataset):
        raise AssertionError('not an instance of xarray.Dataset')

    ds = xr.Dataset({'mesh': xr.DataArray(data='a', attrs=MESH_ATTRS)})

    meta = ugrid.a.attrs

    node_coords = meta['node_coordinates'].split()
    ds.update({'x_of_node': ugrid[node_coords[0]],
               'y_of_node': ugrid[node_coords[1]], })
    ds = ds.rename({ds.x_of_node.dims[0]: 'node'})

    ds.update({'nodes_at_link': ugrid[meta['edge_node_connectivity']]})
    ds = ds.rename({ds.nodes_at_link.dims[0]: 'link',
                    ds.nodes_at_link.dims[1]: 'Two', })

    ds.update({'links_at_patch': ugrid[meta['face_edge_connectivity']]})
    ds = ds.rename({ds.links_at_patch.dims[0]: 'patch',
                    ds.links_at_patch.dims[1]: 'max_patch_links', })

    return ds


def ugrid_as_dataset(grid):
    if isinstance(grid, types.StringTypes):
        ugrid = xr.open_dataset(grid)
    elif isinstance(grid, Graph):
        ugrid = xr.ds
    elif isinstance(grid, dict):
        ugrid = xr.Dataset.from_dict(d)
    elif isinstance(grid, xr.Dataset):
        ugrid = grid
    else:
        raise ValueError(
            'unable to convert {t} to a Dataset'.format(t=type(grid)))

    if not validate_ugrid(ugrid):
        raise ValueError('unable to convert to a ugrid Dataset')
    else:
        return grid


def ugrid_as_dual(ugrid):
    rename = {
        'node': 'corner',
        'link': 'face',
        'patch': 'cell',
        'x_of_node': 'x_of_corner',
        'y_of_node': 'y_of_corner',
        'nodes_at_link': 'corners_at_face',
        'links_at_patch': 'faces_at_cell',
        'max_patch_links': 'max_cell_faces',
    }
    return ugrid.rename(rename)


def ugrid_from_unstructured(node_y_and_x, links=None, patches=None):
    ugrid = xr.Dataset({'mesh': xr.DataArray(data='a', attrs=MESH_ATTRS)})

    update_node_coords(ugrid, node_y_and_x)

    if links is not None:
        update_nodes_at_link(ugrid, links)

    if patches is not None and 'nodes_at_link' in ugrid:
        update_links_at_patch(ugrid, patches)

    return ugrid


def update_node_coords(ugrid, node_y_and_x):
    node_y, node_x = (np.asarray(node_y_and_x[0], dtype=float),
                      np.asarray(node_y_and_x[1], dtype=float))
    y_of_node = xr.DataArray(
        data=node_y.reshape((-1, )),
        coords={'node': np.arange(node_y.size)},
        dims=('node', ),
        attrs={ 'long_name': 'y', 'units': 'm',})
    x_of_node = xr.DataArray(
        data=node_x.reshape((-1, )),
        coords={'node': np.arange(node_x.size)},
        dims=('node', ),
        attrs={ 'long_name': 'x', 'units': 'm',})
    ugrid.update({'y_of_node': y_of_node, 'x_of_node': x_of_node})

    return ugrid


def update_nodes_at_link(ugrid, node_links):
    node_links = np.asarray(node_links, dtype=np.int).reshape((-1, 2))
    nodes_at_link = xr.DataArray(
        data=node_links,
        dims=('link', 'Two'),
        attrs={'cf_role': 'edge_node_connectivity',
               'long_name': 'nodes a link tail and head',
               'start_index': 0})
    ugrid.update({'nodes_at_link': nodes_at_link})


def update_links_at_patch(ugrid, patches):
    from .matrix.at_patch import links_at_patch

    if len(patches) > 0:
        patches = flatten_jagged_array(patches, dtype=int)
    patch_links = links_at_patch(patches)
    links_at_patch = xr.DataArray(
        data=patch_links,
        dims=('patch', 'max_patch_links'),
        attrs={'cf_role': 'face_edge_connectivity',
               'long_name': 'Maps every face to its edges',
               'start_index': 0}
    )
    ugrid.update({'links_at_patch': links_at_patch})


def update_nodes_at_patch(ugrid, nodes_at_patch):
    nodes_at_patch = xr.DataArray(
        data=nodes_at_patch,
        dims=('patch', 'max_patch_links'),
        attrs={'cf_role': 'patch_node_connectivity',
               'long_name': 'nodes defining patches',
               'start_index': 0})
    ugrid.update({'nodes_at_patch': nodes_at_patch})


def update_node_at_cell(ugrid, node_at_cell):
    node_at_cell = xr.DataArray(
        data=node_at_cell,
        dims=('cell', ),
        attrs={'cf_role': 'node_cell_connectivity',
               'long_name': 'node at cell',
               'start_index': 0})
    ugrid.update({'node_at_cell': node_at_cell})


def update_nodes_at_face(ugrid, nodes_at_face):
    nodes_at_face = xr.DataArray(
        data=nodes_at_face,
        dims=('face', 'Two'),
        attrs={'cf_role': 'node_face_connectivity',
               'long_name': 'nodes at face',
               'start_index': 0})
    ugrid.update({'nodes_at_face': nodes_at_face})



import numpy as np


def get_angle_of_link(graph, out=None):
    """Get angles of links in a graph.

    Parameters
    ----------
    graph : `Graph`
        A `Graph`.

    Returns
    -------
    ndarray of float
        Angle of each link as measured in radians from link tail to head.
    out : ndarray of float, optional
        Buffer to place output.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import UniformRectilinearGraph
    >>> from landlab.graph.quantity.of_link import get_angle_of_link

    >>> graph = UniformRectilinearGraph((2, 2))
    >>> get_angle_of_link(graph) * 180. / np.pi
    array([  0.,  90.,  90.,   0.])

    >>> angles = np.empty(graph.number_of_links, dtype=float)
    >>> rtn = get_angle_of_link(graph, out=angles)
    >>> angles is rtn
    True
    """
    if out is None:
        out = np.empty(graph.number_of_links, dtype=float)

    y = graph.y_of_node[graph.nodes_at_link]
    x = graph.x_of_node[graph.nodes_at_link]

    np.arctan2(np.diff(y).flat, np.diff(x).flat, out=out)

    return np.mod(out, 2. * np.pi, out=out)


def get_midpoint_of_link(graph, out=None):
    """Get xy position of the midpoint of links.

    Parameters
    ----------
    graph : `Graph`
        A `Graph`.
    out : ndarray of float, optional
        Buffer to place output.

    Returns
    -------
    ndarray of float, shape `(n_links, 2)`
        Coordinates of link midpoints.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import UniformRectilinearGraph
    >>> from landlab.graph.quantity.of_link import get_midpoint_of_link

    >>> graph = UniformRectilinearGraph((2, 2))
    >>> get_midpoint_of_link(graph)
    array([[ 0.5,  0. ],
           [ 0. ,  0.5],
           [ 1. ,  0.5],
           [ 0.5,  1. ]])

    >>> points = np.empty((graph.number_of_links, 2), dtype=float)
    >>> rtn = get_midpoint_of_link(graph, out=points)
    >>> points is rtn
    True
    """
    from .ext.of_link import calc_midpoint_of_link

    if out is None:
        out = np.empty((graph.number_of_links, 2), dtype=float)

    calc_midpoint_of_link(graph.nodes_at_link, graph.x_of_node,
                          graph.y_of_node, out)

    return out


def get_length_of_link(graph):
    nodes_at_link = graph.nodes_at_link
    dx = (graph.x_of_node[nodes_at_link[:, 0]] -
          graph.x_of_node[nodes_at_link[:, 1]])
    dy = (graph.y_of_node[nodes_at_link[:, 0]] -
          graph.y_of_node[nodes_at_link[:, 1]])
    return np.sqrt(dx ** 2 + dy ** 2)



import numpy as np
cimport numpy as np
cimport cython

from libc.stdlib cimport malloc, free


cdef extern from "math.h":
    double atan2(double y, double x) nogil


from .spoke_sort import sort_spokes_at_wheel
from .argsort cimport argsort_int

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def reverse_one_to_one(np.ndarray[DTYPE_t, ndim=1] mapping,
                       np.ndarray[DTYPE_t, ndim=1] out):
    cdef int n_elements = mapping.size
    cdef int index
    cdef int id_

    for index in range(n_elements):
        id_ = mapping[index]
        if id_ >= 0:
          out[id_] = index


@cython.boundscheck(False)
def reverse_one_to_many(np.ndarray[DTYPE_t, ndim=2] mapping,
                        np.ndarray[DTYPE_t, ndim=2] out):
    cdef int n_elements = mapping.shape[0]
    cdef int n_cols = mapping.shape[1]
    cdef int out_rows = out.shape[0]
    cdef int index
    cdef int id_
    cdef int *count = <int *>malloc(out_rows * sizeof(int))

    try:
        for index in range(out_rows):
            count[index] = 0

        for index in range(n_elements):
            for col in range(n_cols):
                id_ = mapping[index, col]
                if id_ >= 0:
                    out[id_, count[id_]] = index
                    count[id_] += 1
    finally:
        free(count)


@cython.boundscheck(False)
def remap_graph_element(np.ndarray[DTYPE_t, ndim=1] elements,
                        np.ndarray[DTYPE_t, ndim=1] old_to_new):
    """Remap elements in an array in place.

    Parameters
    ----------
    elements : ndarray of int
        Identifiers of elements.
    old_to_new : ndarray of int
        Mapping from the old identifier to the new identifier.
    """
    cdef int n_elements = elements.size
    cdef int i

    for i in range(n_elements):
      elements[i] = old_to_new[elements[i]]


@cython.boundscheck(False)
def remap_graph_element_ignore(np.ndarray[DTYPE_t, ndim=1] elements,
                               np.ndarray[DTYPE_t, ndim=1] old_to_new,
                               DTYPE_t bad_val):
    """Remap elements in an array in place, ignoring bad values.

    Parameters
    ----------
    elements : ndarray of int
        Identifiers of elements.
    old_to_new : ndarray of int
        Mapping from the old identifier to the new identifier.
    bad_val : int
        Ignore values in the input array when remapping.
    """
    cdef int n_elements = elements.size
    cdef int i

    for i in range(n_elements):
        if elements[i] != bad_val:
            elements[i] = old_to_new[elements[i]]


@cython.boundscheck(False)
def reorder_patches(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                    np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                    np.ndarray[DTYPE_t, ndim=1] sorted_patches):
    cdef int i
    cdef int patch
    cdef int offset
    cdef int n_links
    cdef int n_patches = len(sorted_patches)
    cdef int *new_offset = <int *>malloc(len(offset_to_patch) * sizeof(int))
    cdef int *new_patches = <int *>malloc(len(links_at_patch) * sizeof(int))

    try:
        new_offset[0] = 0
        for patch in range(n_patches):
            offset = offset_to_patch[sorted_patches[patch]]
            n_links = offset_to_patch[sorted_patches[patch] + 1] - offset

            new_offset[patch + 1] = new_offset[patch] + n_links
            for i in range(n_links):
                new_patches[new_offset[patch] + i] = links_at_patch[offset + i]

        for i in range(len(links_at_patch)):
            links_at_patch[i] = new_patches[i]
        for i in range(len(offset_to_patch)):
            offset_to_patch[i] = new_offset[i]
    finally:
        free(new_offset)
        free(new_patches)


@cython.boundscheck(False)
def calc_center_of_patch(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                         np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                         np.ndarray[np.float_t, ndim=2] xy_at_link,
                         np.ndarray[np.float_t, ndim=2] xy_at_patch):
    cdef int patch
    cdef int link
    cdef int i
    cdef int offset
    cdef int n_links
    cdef int n_patches = len(xy_at_patch)
    cdef float x
    cdef float y

    for patch in range(n_patches):
        offset = offset_to_patch[patch]
        n_links = offset_to_patch[patch + 1] - offset
        x = 0.
        y = 0.
        for i in range(offset, offset + n_links):
            link = links_at_patch[i]
            x += xy_at_link[link, 0]
            y += xy_at_link[link, 1]
        xy_at_patch[patch, 0] = x / n_links
        xy_at_patch[patch, 1] = y / n_links


@cython.boundscheck(False)
def reorder_links_at_patch(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                           np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                           np.ndarray[np.float_t, ndim=2] xy_of_link):
    cdef int n_patches = len(offset_to_patch) - 1

    xy_of_patch = np.empty((n_patches, 2), dtype=float)
    calc_center_of_patch(links_at_patch, offset_to_patch, xy_of_link,
                         xy_of_patch)
    sort_spokes_at_wheel(links_at_patch, offset_to_patch, xy_of_patch,
                         xy_of_link)


cdef _argsort_links(long * links, int n_links, long * nodes, long * ordered):
    cdef int n_nodes = 2 * n_links
    cdef int * index = <int *>malloc(n_nodes * sizeof(int))

    try:
        argsort_int(nodes, n_nodes, index)

        i = 0
        for link in range(n_links):
            print index[i]
            # if nodes[index[i]] != nodes[index[i + 1]]:
            #     raise ValueError('open patch')

            ordered[i / 2] = index[i] / 2
            i += 2

            # if nodes[index[i]] == nodes[index[i - 1]]:
            #     raise ValueError('triple-point')
    finally:
        free(index)


@cython.boundscheck(False)
def connect_links(np.ndarray[long, ndim=1, mode="c"] links,
                  np.ndarray[long, ndim=2] nodes_at_link):
    cdef long n_links = links.size
    cdef long * nodes = <long *>malloc(2 * n_links * sizeof(long))
    cdef long * ordered = <long *>malloc(n_links * sizeof(long))
    cdef long * buff = <long *>malloc(n_links * sizeof(long))
    cdef long node
    cdef long link

    try:
        node = 0
        for link in range(n_links):
            nodes[node] = nodes_at_link[link, 0]
            nodes[node + 1] = nodes_at_link[link, 1]
            node += 2

        _argsort_links(&links[0], n_links, nodes, ordered)

        for link in range(n_links):
            buff[link] = links[ordered[link]]

        for link in range(n_links):
            links[link] = buff[link]

    finally:
        free(buff)
        free(ordered)
        free(nodes)


cdef reverse_order(long * array, long size):
    cdef long i
    cdef long temp

    for i in range(size / 2):
        temp = array[i]
        array[i] = array[(size - 1) - i]
        array[(size - 1) - i] = temp

        # array[i], array[size - 1] = array[size - 1], array[i]


@cython.boundscheck(False)
def reverse_element_order(np.ndarray[long, ndim=2] links_at_patch,
                          np.ndarray[long, ndim=1] patches):
    cdef long n_patches = patches.shape[0]
    cdef long max_links = links_at_patch.shape[1]
    cdef long patch
    cdef long n
    cdef long i

    for i in range(n_patches):
        patch = patches[i]
        # for n in range(1, max_links):
        #     if links_at_patch[patch, n] == -1:
        #         break
        # reverse_order(&links_at_patch[patch, 1], n - 1)

        n = 1
        while n < max_links:
            if links_at_patch[patch, n] == -1:
                break
            n += 1
        reverse_order(&links_at_patch[patch, 1], n - 1)

@cython.boundscheck(False)
def get_angle_of_link(np.ndarray[DTYPE_t, ndim=2] nodes_at_link,
                      np.ndarray[np.float_t, ndim=2] xy_of_node,
                      np.ndarray[np.float_t, ndim=1] angle_of_link):
    cdef int link
    cdef float link_tail_x
    cdef float link_tail_y
    cdef float link_head_x
    cdef float link_head_y
    cdef int n_links = nodes_at_link.shape[0]

    for link in range(n_links):
        link_tail_x = xy_of_node[nodes_at_link[link][0]][0]
        link_tail_y = xy_of_node[nodes_at_link[link][0]][1]
        link_head_x = xy_of_node[nodes_at_link[link][1]][0]
        link_head_y = xy_of_node[nodes_at_link[link][1]][1]

        angle_of_link[link] = atan2(link_head_y - link_tail_y,
                                    link_head_x - link_head_y)


@cython.boundscheck(False)
def reorient_links(np.ndarray[DTYPE_t, ndim=2] nodes_at_link,
                   np.ndarray[DTYPE_t, ndim=1] xy_of_node):
    """Reorient links to point up and to the right.

    Parameters
    ----------
    nodes_at_link : ndarray of int, shape `(n_nodes, 2)`
        Identifier for node at link tail and head.
    xy_of_node : ndarray of float, shape `(n_nodes, 2)`
        Coordinate of node as `(x, y)`.
    """
    cdef int link
    cdef int temp
    cdef double angle
    cdef int n_links = nodes_at_link.shape[0]
    cdef double minus_45 = - np.pi * .25
    cdef double plus_135 = np.pi * .75

    angle_of_link = np.empty(n_links, dtype=n_links)
    get_angle_of_link(nodes_at_link, xy_of_node, angle_of_link)

    for link in range(n_links):
        angle = angle_of_link[link]
        if angle < minus_45 or angle > plus_135:
            temp = nodes_at_link[link, 0]
            nodes_at_link[link, 0] = nodes_at_link[link, 1]
            nodes_at_link[link, 1] = temp



"""Landlab component that simulates overland flow.

This component simulates overland flow using the 2-D numerical model of
shallow-water flow over topography using the de Almeida et al., 2012
algorithm for storage-cell inundation modeling.

.. codeauthor:: Jordan Adams

Examples
--------
>>> import numpy as np
>>> from landlab import RasterModelGrid
>>> from landlab.components.overland_flow import OverlandFlow

Create a grid on which to calculate overland flow.

>>> grid = RasterModelGrid((4, 5))

The grid will need some data to provide the overland flow component. To
check the names of the fields that provide input to the overland flow
component use the *input_var_names* class property.

>>> OverlandFlow.input_var_names
('surface_water__depth', 'topographic__elevation')

Create fields of data for each of these input variables.

>>> grid.at_node['topographic__elevation'] = np.array([
...     0., 0., 0., 0., 0.,
...     1., 1., 1., 1., 1.,
...     2., 2., 2., 2., 2.,
...     3., 3., 3., 3., 3.])
>>> grid.at_node['surface_water__depth'] = np.array([
...     0. , 0. , 0. , 0. , 0. ,
...     0. , 0. , 0. , 0. , 0. ,
...     0. , 0. , 0. , 0. , 0. ,
...     0.1, 0.1, 0.1, 0.1, 0.1])

Instantiate the `OverlandFlow` component to work on this grid, and run it.

>>> of = OverlandFlow(grid, steep_slopes=True)
>>> of.run_one_step()

After calculating the overland flow, new fields have been added to the
grid. Use the *output_var_names* property to see the names of the fields that
have been changed.

>>> of.output_var_names
('surface_water__depth', 'surface_water__discharge', 'water_surface__gradient')

The `surface_water__depth` field is defined at nodes.

>>> of.var_loc('surface_water__depth')
'node'
>>> grid.at_node['surface_water__depth'] # doctest: +NORMALIZE_WHITESPACE
array([  1.00000000e-05,   1.00000000e-05,   1.00000000e-05,
         1.00000000e-05,   1.00000000e-05,   1.00000000e-05,
         1.00000000e-05,   1.00000000e-05,   1.00000000e-05,
         1.00000000e-05,   1.00000000e-05,   2.00100000e-02,
         2.00100000e-02,   2.00100000e-02,   1.00000000e-05,
         1.00010000e-01,   1.00010000e-01,   1.00010000e-01,
         1.00010000e-01,   1.00010000e-01])

The `surface_water__discharge` field is defined at links. Because our initial
topography was a dipping plane, there is no water discharge in the horizontal
direction, only toward the bottom of the grid.

>>> of.var_loc('surface_water__discharge')
'link'
>>> q = grid.at_link['surface_water__discharge'] # doctest: +NORMALIZE_WHITESPACE
>>> np.all(q[grid.horizontal_links] == 0.)
True
>>> np.all(q[grid.vertical_links] <= 0.)
True

The *water_surface__gradient* is also defined at links.

>>> of.var_loc('water_surface__gradient')
'link'
>>> grid.at_link['water_surface__gradient'] # doctest: +NORMALIZE_WHITESPACE
array([ 0. ,  0. ,  0. ,  0. ,
        0. ,  1. ,  1. ,  1. ,  0. ,
        0. ,  0. ,  0. ,  0. ,
        0. ,  1. ,  1. ,  1. ,  0. ,
        0. ,  0. ,  0. ,  0. ,
        0. ,  1.1,  1.1,  1.1,  0. ,
        0. ,  0. ,  0. ,  0. ])
"""
from landlab import Component, FieldError
import numpy as np
from landlab.grid.structured_quad import links
from landlab.utils.decorators import use_file_name_or_kwds


_SEVEN_OVER_THREE = 7.0 / 3.0


class OverlandFlow(Component):

    """Simulate overland flow using de Almeida approximations.

    Landlab component that simulates overland flow using the de Almeida
    et al., 2012 approximations of the 1D shallow water equations to be used
    for 2D flood inundation modeling.

    This component calculates discharge, depth and shear stress after some
    precipitation event across any raster grid. Default input file is named
    "overland_flow_input.txt' and is contained in the
    landlab.components.overland_flow folder.

    The primary method of this class is :func:`run_one_step`.
    """

    _name = 'OverlandFlow'

    _cite_as = """@article{adams2017landlab,
	title={The Landlab v1. 0 OverlandFlow component: a Python
            tool for computing shallow-water flow across watersheds},
        author={Adams, Jordan M and Gasparini, Nicole M and
            Hobley, Daniel EJ and Tucker, Gregory E and
            Hutton, Eric WH and Nudurupati, Sai S and
            Istanbulluoglu, Erkan},
        journal={Geoscientific Model Development},
        volume={10},
        number={4},
        pages={1645},
        year={2017},
        publisher={Copernicus GmbH}
        }
    """

    _input_var_names = (
        'surface_water__depth',
        'topographic__elevation',
    )

    _output_var_names = (
        'surface_water__depth',
        'surface_water__discharge',
        'water_surface__gradient',
    )

    _var_units = {
        'surface_water__depth': 'm',
        'surface_water__discharge': 'm3/s',
        'topographic__elevation': 'm',
        'water_surface__gradient': '-',
    }

    _var_mapping = {
        'surface_water__depth': 'node',
        'topographic__elevtation': 'node',
        'surface_water__discharge': 'link',
        'water_surface__gradient': 'link',
    }

    _var_doc = {
        'surface_water__depth': 'The depth of water at each node.',
        'topographic__elevtation': 'The land surface elevation.',
        'surface_water__discharge': 'The discharge of water on active links.',
        'water_surface__gradient': 'Downstream gradient of the water surface.',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, default_fixed_links=False, h_init=0.00001,
                 alpha=0.7, mannings_n=0.03, g=9.81, theta=0.8,
                 rainfall_intensity=0.0, steep_slopes=False, **kwds):
        """Create an overland flow component.

        Parameters
        ----------
        grid : RasterModelGrid
            A landlab grid.
        h_init : float, optional
            Thicknes of initial thin layer of water to prevent divide by zero
            errors (m).
        alpha : float, optional
            Time step coeffcient, described in Bates et al., 2010 and
            de Almeida et al., 2012.
        mannings_n : float, optional
            Manning's roughness coefficient.
        g : float, optional
            Acceleration due to gravity (m/s^2).
        theta : float, optional
            Weighting factor from de Almeida et al., 2012.
        rainfall_intensity : float, optional
            Rainfall intensity.
        """
        super(OverlandFlow, self).__init__(grid, **kwds)

        # First we copy our grid
        self._grid = grid

        self.h_init = h_init
        self.alpha = alpha
        
        if type(mannings_n) is str:
            self.mannings_n = self.grid.at_link[mannings_n]
        else:
            self.mannings_n = mannings_n

        self.g = g
        self.theta = theta
        self.rainfall_intensity = rainfall_intensity
        self.steep_slopes = steep_slopes


        # Now setting up fields at the links...
        # For water discharge
        try:
            self.q = grid.add_zeros('surface_water__discharge', at='link',
                units=self._var_units['surface_water__discharge'])

        except FieldError:
            # Field was already set; still, fill it with zeros
            self.q = grid.at_link['surface_water__discharge']
            self.q.fill(0.)

        # For water depths calculated at links
        try:
            self.h_links = grid.add_zeros('surface_water__depth', at='link',
                                          units=self._var_units[
                                              'surface_water__depth'])
        except FieldError:
            self.h_links = grid.at_link['surface_water__depth']
            self.h_links.fill(0.)

        self.h_links += self.h_init

        try:
            self.h = grid.add_zeros('surface_water__depth', at='node',
                units=self._var_units['surface_water__depth'])

        except FieldError:
            # Field was already set
            self.h = grid.at_node['surface_water__depth']

        self.h += self.h_init

        # For water surface slopes at links
        try:
            self.water_surface_slope = grid.add_zeros('water_surface__gradient', at='link')
        except FieldError:
            self.water_surface_slope = grid.at_link['water_surface__gradient']
            self.water_surface_slope.fill(0.)

        # Start time of simulation is at 1.0 s
        self.elapsed_time = 1.0

        self.dt = None
        self.dhdt = grid.zeros()

        # When we instantiate the class we recognize that neighbors have not
        # been found. After the user either calls self.set_up_neighbor_array
        # or self.overland_flow this will be set to True. This is done so
        # that every iteration of self.overland_flow does NOT need to
        # reinitalize the neighbors and saves computation time.
        self.neighbor_flag = False

        # When looking for neighbors, we automatically ignore inactive links
        # by default. However, what about when we want to look at fixed links
        # too? By default, we ignore these, but if they are important to your
        # model and will be updated in your driver loop, they can be used by
        # setting the flag in the initialization of  the class to 'True'
        self.default_fixed_links = default_fixed_links

        # Assiging a class variable to the elevation field.
        self.z = self._grid.at_node['topographic__elevation']

    def calc_time_step(self):
        """Calculate time step.

        Adaptive time stepper from Bates et al., 2010 and de Almeida
        et al., 2012
        """
        self.dt = (self.alpha * self._grid.dx / np.sqrt(self.g * np.amax(
            self._grid.at_node['surface_water__depth'])))

        return self.dt

    def set_up_neighbor_arrays(self):
        """Create and initialize link neighbor arrays.

        Set up arrays of neighboring horizontal and vertical links that are
        needed for the de Almeida solution.
        """
        # First we identify all active links
        
        self.active_ids = links.active_link_ids(self.grid.shape,
                                                self.grid.status_at_node)

        self.active_links_at_open_bdy =self.grid._active_links_at_node(
                                    self.grid.open_boundary_nodes).transpose()
        
        self.active_links_at_open_bdy = self.active_links_at_open_bdy[
                                np.where(self.active_links_at_open_bdy > -1)]

        # And then find all horizontal link IDs (Active and Inactive)
        self.horizontal_ids = links.horizontal_link_ids(self.grid.shape)

        # And make the array 1-D
        self.horizontal_ids = self.horizontal_ids.flatten()

        # Find all horizontal active link ids
        self.horizontal_active_link_ids = links.horizontal_active_link_ids(
            self.grid.shape, self.active_ids)

        # Now we repeat this process for the vertical links.
        # First find the vertical link ids and reshape it into a 1-D array
        self.vertical_ids = links.vertical_link_ids(self.grid.shape).flatten()

        # Find the *active* verical link ids
        self.vertical_active_link_ids = links.vertical_active_link_ids(
            self.grid.shape, self.active_ids)

        if self.default_fixed_links is True:
            fixed_link_ids = links.fixed_link_ids(
                self.grid.shape, self.grid.status_at_node)
            fixed_horizontal_links = links.horizontal_fixed_link_ids(
                self.grid.shape, fixed_link_ids)
            fixed_vertical_links = links.vertical_fixed_link_ids(
                self.grid.shape, fixed_link_ids)
            self.horizontal_active_link_ids = np.maximum(
                self.horizontal_active_link_ids, fixed_horizontal_links)
            self.vertical_active_link_ids = np.maximum(
                self.vertical_active_link_ids, fixed_vertical_links)
            self.active_neighbors = find_active_neighbors_for_fixed_links(
                self.grid)
            
        self.vert_bdy_ids = self.active_links_at_open_bdy[
                            links.is_vertical_link(self.grid.shape, 
                                            self.active_links_at_open_bdy)]
                                                   
        self.vert_bdy_ids = links.nth_vertical_link(self.grid.shape, 
                                                    self.vert_bdy_ids)
        
        self.horiz_bdy_ids = self.active_links_at_open_bdy[
                            links.is_horizontal_link(self.grid.shape, 
                                            self.active_links_at_open_bdy)]
                                                     
        self.horiz_bdy_ids = links.nth_horizontal_link(self.grid.shape,
                                                  self.horiz_bdy_ids)
        
        # Using the active vertical link ids we can find the north
        # and south vertical neighbors
        self.north_neighbors = links.vertical_north_link_neighbor(
            self.grid.shape, self.vertical_active_link_ids)
        self.south_neighbors = links.vertical_south_link_neighbor(
            self.grid.shape, self.vertical_active_link_ids)

        # Using the horizontal active link ids, we can find the west and
        # east neighbors
        self.west_neighbors = links.horizontal_west_link_neighbor(
            self.grid.shape, self.horizontal_active_link_ids)
        self.east_neighbors = links.horizontal_east_link_neighbor(
            self.grid.shape, self.horizontal_active_link_ids)

        ## replace bdy condition links
        (ids, ) = np.where(self.west_neighbors[self.horiz_bdy_ids] == -1)
        ids = self.horiz_bdy_ids[ids]
        self.west_neighbors[ids] = self.horizontal_active_link_ids[ids]
        
        (ids, ) = np.where(self.east_neighbors[self.horiz_bdy_ids] == -1)
        ids = self.horiz_bdy_ids[ids]
        self.east_neighbors[ids] = self.horizontal_active_link_ids[ids]

        (ids, ) = np.where(self.north_neighbors[self.vert_bdy_ids] == -1)
        ids = self.vert_bdy_ids[ids]
        self.north_neighbors[ids] = self.vertical_active_link_ids[ids]
        
        (ids, ) = np.where(self.south_neighbors[self.vert_bdy_ids] == -1)
        ids = self.vert_bdy_ids[ids]
        self.south_neighbors[ids] = self.vertical_active_link_ids[ids]
        
        # Set up arrays for discharge in the horizontal & vertical directions.
        self.q_horizontal = np.zeros(links.number_of_horizontal_links(
            self.grid.shape))
        self.q_vertical = np.zeros(links.number_of_vertical_links(
            self.grid.shape))

        # Once the neighbor arrays are set up, we change the flag to True!
        self.neighbor_flag = True

    def overland_flow(self, dt=None):
        """Generate overland flow across a grid.

        For one time step, this generates 'overland flow' across a given grid
        by calculating discharge at each node.

        Using the depth slope product, shear stress is calculated at every
        node.

        Outputs water depth, discharge and shear stress values through time at
        every point in the input grid.
        """
        # DH adds a loop to enable an imposed tstep while maintaining stability
        local_elapsed_time = 0.
        if dt is None:
            dt = np.inf  # to allow the loop to begin
        while local_elapsed_time < dt:
            dt_local = self.calc_time_step()
            # Can really get into trouble if nothing happens but we still run:
            if not dt_local < np.inf:
                break
            if local_elapsed_time + dt_local > dt:
                dt_local = dt - local_elapsed_time
            self.dt = dt_local

            # First, we check and see if the neighbor arrays have been
            # initialized
            if self.neighbor_flag is False:
                self.set_up_neighbor_arrays()

            # In case another component has added data to the fields, we just
            # reset our water depths, topographic elevations and water
            # discharge variables to the fields.
            self.h = self.grid['node']['surface_water__depth']
            self.z = self.grid['node']['topographic__elevation']
            self.q = self.grid['link']['surface_water__discharge']
            self.h_links = self.grid['link']['surface_water__depth']
         
            # Here we identify the core nodes and active links for later use.
            self.core_nodes = self.grid.core_nodes
            self.active_links = self.grid.active_links

            # Per Bates et al., 2010, this solution needs to find difference
            # between the highest water surface in the two cells and the
            # highest bed elevation
            zmax = self._grid.map_max_of_link_nodes_to_link(self.z)
            w = self.h + self.z
            wmax = self._grid.map_max_of_link_nodes_to_link(w)
            hflow = wmax[self._grid.active_links] - zmax[
                self._grid.active_links]

            # Insert this water depth into an array of water depths at the
            # links.
            self.h_links[self.active_links] = hflow

            # Now we calculate the slope of the water surface elevation at
            # active links
            self.water_surface__gradient = (
                self.grid.calc_grad_at_link(w)[self.grid.active_links])

            # And insert these values into an array of all links
            self.water_surface_slope[self.active_links
                                      ] = self.water_surface__gradient
            # If the user chooses to set boundary links to the neighbor value,
            # we set the discharge array to have the boundary links set to
            # their neighbor value
            if self.default_fixed_links is True:
                self.q[self.grid.fixed_links] = self.q[self.active_neighbors]

            # Now we can calculate discharge. To handle links with neighbors
            # that do not exist, we will do a fancy indexing trick. Non-
            # existent links or inactive links have an index of '-1', which in
            # Python, looks to the end of a list or array. To accommodate these
            # '-1' indices, we will simply insert an value of 0.0 discharge (in
            # units of L^2/T) to the end of the discharge array.
            self.q = np.append(self.q, [0])
            
            horiz = self.horizontal_ids
            vert = self.vertical_ids
            # Now we calculate discharge in the horizontal direction
            try:
                self.q[horiz] = ((
                    self.theta * self.q[horiz] + (1. - self.theta) / 2. *
                    (self.q[self.west_neighbors] + self.q[self.east_neighbors]) -
                    self.g * self.h_links[horiz] * self.dt *
                    self.water_surface_slope[horiz]) / (1 + self.g * self.dt 
                    * self.mannings_n ** 2. * abs(self.q[horiz]) / 
                    self.h_links[horiz] ** _SEVEN_OVER_THREE))
    
                # ... and in the vertical direction
                self.q[vert] = ((
                    self.theta * self.q[vert] + (1 - self.theta) / 2. *
                    (self.q[self.north_neighbors] + self.q[self.south_neighbors]) -
                    self.g * self.h_links[vert] * self.dt *
                    self.water_surface_slope[vert]) / (1 + self.g * self.dt * 
                    self.mannings_n ** 2. * abs(self.q[vert]) / 
                    self.h_links[vert] ** _SEVEN_OVER_THREE))
                
            except ValueError:
                self.mannings_n = self.grid['link']['mannings_n']
                # if manning's n in a field
                # calc discharge in horizontal
                self.q[horiz] = ((
                    self.theta * self.q[horiz] + (1. - self.theta) / 2. *
                    (self.q[self.west_neighbors] + self.q[self.east_neighbors]) -
                    self.g * self.h_links[horiz] * self.dt *
                    self.water_surface_slope[horiz]) / (1 + self.g * self.dt 
                    * self.mannings_n[horiz]** 2. * abs(self.q[horiz]) / 
                    self.h_links[horiz] ** _SEVEN_OVER_THREE))
    
                # ... and in the vertical direction
                self.q[vert] = ((
                    self.theta * self.q[vert] + (1 - self.theta) / 2. *
                    (self.q[self.north_neighbors] + self.q[self.south_neighbors]) -
                    self.g * self.h_links[vert] * self.dt *
                    self.water_surface_slope[self.vertical_ids]) / (1 + self.g 
                    * self.dt * self.mannings_n[vert] ** 2. * abs(self.q[vert])
                    / self.h_links[vert] ** _SEVEN_OVER_THREE))

            # Now to return the array to its original length (length of number
            # of all links), we delete the extra 0.0 value from the end of the
            # array.
            self.q = np.delete(self.q, len(self.q) - 1)

            # Updating the discharge array to have the boundary links set to
            # their neighbor
            if self.default_fixed_links is True:
                self.q[self.grid.fixed_links] = self.q[self.active_neighbors]

            if self.steep_slopes is True:
                # To prevent water from draining too fast for our time steps...
                # Our Froude number.
                Fr = 1.0
                # Our two limiting factors, the froude number and courant
                # number.
                # Looking a calculated q to be compared to our Fr number.
                calculated_q = (self.q / self.h_links) / np.sqrt(self.g *
                                                                 self.h_links)

                # Looking at our calculated q and comparing it to Courant no.,
                q_courant = self.q * self.dt / self.grid.dx

                # Water depth split equally between four links..
                water_div_4 = self.h_links / 4.

                # IDs where water discharge is positive...
                (positive_q, ) = np.where(self.q > 0)

                # ... and negative.
                (negative_q, ) = np.where(self.q < 0)

                # Where does our calculated q exceed the Froude number? If q
                # does exceed the Froude number, we are getting supercritical
                # flow and discharge needs to be reduced to maintain stability.
                (Froude_logical, ) = np.where((calculated_q) > Fr)
                (Froude_abs_logical, ) = np.where(abs(calculated_q) > Fr)

                # Where does our calculated q exceed the Courant number and
                # water depth divided amongst 4 links? If the calculated q
                # exceeds the Courant number and is greater than the water
                # depth divided by 4 links, we reduce discharge to maintain
                # stability.
                (water_logical, ) = np.where(q_courant > water_div_4)
                (water_abs_logical, ) = np.where(abs(q_courant) > water_div_4)

                # Where are these conditions met? For positive and negative q,
                # there are specific rules to reduce q. This step finds where
                # the discharge values are positive or negative and where
                # discharge exceeds the Froude or Courant number.
                self.if_statement_1 = np.intersect1d(positive_q,
                                                     Froude_logical)
                self.if_statement_2 = np.intersect1d(negative_q,
                                                     Froude_abs_logical)
                self.if_statement_3 = np.intersect1d(positive_q, water_logical)
                self.if_statement_4 = np.intersect1d(negative_q,
                                                     water_abs_logical)

                # Rules 1 and 2 reduce discharge by the Froude number.
                self.q[self.if_statement_1] = (
                    self.h_links[self.if_statement_1] *
                    (np.sqrt(self.g * self.h_links[self.if_statement_1]) * Fr))

                self.q[self.if_statement_2] = (
                    0. - (self.h_links[self.if_statement_2] *
                          np.sqrt(self.g * self.h_links[self.if_statement_2]) *
                          Fr))

                # Rules 3 and 4 reduce discharge by the Courant number.
                self.q[self.if_statement_3] = (((
                    self.h_links[self.if_statement_3] * self.grid.dx) / 5.) /
                    self.dt)

                self.q[self.if_statement_4] = (0. - (self.h_links[
                    self.if_statement_4] * self.grid.dx / 5.) / self.dt)

            # Once stability has been restored, we calculate the change in
            # water depths on all core nodes by finding the difference between
            # inputs (rainfall) and the inputs/outputs (flux divergence of
            # discharge)
            self.dhdt = (self.rainfall_intensity -
                         self.grid.calc_flux_div_at_node(self.q))

            # Updating our water depths...
            self.h[self.core_nodes] = (self.h[self.core_nodes] +
                                       self.dhdt[self.core_nodes] * self.dt)

            # To prevent divide by zero errors, a minimum threshold water depth
            # must be maintained. To reduce mass imbalances, this is set to
            # find locations where water depth is smaller than h_init (default
            # is 0.001) and the new value is self.h_init * 10^-3. This was set
            # as it showed the smallest amount of mass creation in the grid
            # during testing.
            if self.steep_slopes is True:
                self.h[self.h < self.h_init] = self.h_init * 10.0 ** -3

            # And reset our field values with the newest water depth and
            # discharge.
            self.grid.at_node['surface_water__depth'] = self.h
            self.grid.at_link['surface_water__discharge'] = self.q
#
#
#            self.helper_q = self.grid.map_upwind_node_link_max_to_node(self.q)
#            self.helper_s = self.grid.map_upwind_node_link_max_to_node(
#                                                    self.water_surface_slope)
#            
#            self.helper_q = self.grid.map_max_of_link_nodes_to_link(self.helper_q)
#            self.helper_s = self.grid.map_max_of_link_nodes_to_link(self.helper_s)
#
#            self.grid['link']['surface_water__discharge'][
#                     self.active_links_at_open_bdy] = self.helper_q[
#                     self.active_links_at_open_bdy]
#
#            self.grid['link']['water_surface__gradient'][
#                self.active_links_at_open_bdy] = self.helper_s[
#                self.active_links_at_open_bdy]
            ## Update nodes near boundary locations - nodes adjacent to
            ## boundaries may have discharge and water surface slopes
            ## artifically reduced due to boundary effects. This step removes
            ## those errors. 
            
            if dt is np.inf:
                break
            local_elapsed_time += self.dt

    def run_one_step(self, dt=None):
        """Generate overland flow across a grid.

        For one time step, this generates 'overland flow' across a given grid
        by calculating discharge at each node.

        Using the depth slope product, shear stress is calculated at every
        node.

        Outputs water depth, discharge and shear stress values through time at
        every point in the input grid.
        """
        self.overland_flow(dt=dt)


    def discharge_mapper(self, input_discharge, convert_to_volume=False):
        """
        Maps discharge value from links onto nodes.

        This method takes the discharge values on links and determines the
        links that are flowing INTO a given node. The fluxes moving INTO a
        given node are summed.

        This method ignores all flow moving OUT of a given node.

        This takes values from the OverlandFlow component (by default) in
        units of [L^2/T]. If the convert_to_cms flag is raised as True, this
        method converts discharge to units [L^3/T] - as of Aug 2016, only
        operates for square RasterModelGrid instances.

        The output array is of length grid.number_of_nodes and can be used
        with the Landlab imshow_grid plotter.

        Returns a numpy array (discharge_vals)
        """

        discharge_vals = np.zeros(self.grid.number_of_links)
        discharge_vals[:] = input_discharge[:]

        if convert_to_volume == True:
            discharge_vals *= self.grid.dx
        else:
            pass

        discharge_vals = (discharge_vals[self.grid.links_at_node] *
                                        self.grid.link_dirs_at_node)

        discharge_vals = discharge_vals.flatten()

        discharge_vals[np.where(discharge_vals < 0)] = 0.0

        discharge_vals = discharge_vals.reshape(self.grid.number_of_nodes, 4)

        discharge_vals = discharge_vals.sum(axis=1)

        return discharge_vals


def find_active_neighbors_for_fixed_links(grid):
    """Find active link neighbors for every fixed link.

    Specialized link ID function used to ID the active links that neighbor
    fixed links in the vertical and horizontal directions.

    If the user wants to assign fixed gradients or values to the fixed
    links dynamically, this function identifies the nearest active_link
    neighbor.

    Each fixed link can either have 0 or 1 active neighbor. This function
    finds if and where that active neighbor is and stores those IDs in
    an array.

    Parameters
    ----------
    grid : RasterModelGrid
        A landlab grid.

    Returns
    -------
    ndarray of int, shape `(*, )`
        Flat array of links.


    Examples
    --------
    >>> from landlab.grid.structured_quad.links import neighbors_at_link
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow.generate_overland_flow_deAlmeida import find_active_neighbors_for_fixed_links

    >>> from landlab import RasterModelGrid, FIXED_GRADIENT_BOUNDARY

    >>> grid = RasterModelGrid((4, 5))
    >>> grid.status_at_node[:5] = FIXED_GRADIENT_BOUNDARY
    >>> grid.status_at_node[::5] = FIXED_GRADIENT_BOUNDARY
    >>> grid.status_at_node # doctest: +NORMALIZE_WHITESPACE
    array([2, 2, 2, 2, 2,
           2, 0, 0, 0, 1,
           2, 0, 0, 0, 1,
           2, 1, 1, 1, 1], dtype=uint8)

    >>> grid.fixed_links
    array([ 5,  6,  7,  9, 18])
    >>> grid.active_links
    array([10, 11, 12, 14, 15, 16, 19, 20, 21, 23, 24, 25])

    >>> find_active_neighbors_for_fixed_links(grid)
    array([14, 15, 16, 10, 19])

    >>> rmg = RasterModelGrid((4, 7))

    >>> rmg.at_node['topographic__elevation'] = rmg.zeros(at='node')
    >>> rmg.at_link['topographic__slope'] = rmg.zeros(at='link')

    >>> rmg.set_fixed_link_boundaries_at_grid_edges(True, True, True, True)
    >>> find_active_neighbors_for_fixed_links(rmg)
    array([20, 21, 22, 23, 24, 14, 17, 27, 30, 20, 21, 22, 23, 24])
    """
    neighbors = links.neighbors_at_link(grid.shape, grid.fixed_links).flat
    return neighbors[np.in1d(neighbors, grid.active_links)]



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_boundary_node():
    rmg = RasterModelGrid(5, 6)
    assert rmg.node_has_boundary_neighbor(0)
    assert not rmg.node_has_boundary_neighbor(14)


def test_last_index():
    rmg = RasterModelGrid((4, 5))
    assert rmg.node_has_boundary_neighbor(-1)


def test_id_as_list():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.node_has_boundary_neighbor([-1, 0]),
                       np.array([True, True]))


def test_id_as_array():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.node_has_boundary_neighbor(np.arange(20)),
                       np.array([True, True, True, True, True,
                                 True, True, True, True, True,
                                 True, True, True, True, True,
                                 True, True, True, True, True]))


def test_id_as_array_with_one_interior():
    rmg = RasterModelGrid(5, 5)
    assert_array_equal(rmg.node_has_boundary_neighbor(np.arange(25)),
                       np.array([True, True,  True, True, True,
                                 True, True,  True, True, True,
                                 True, True, False, True, True,
                                 True, True,  True, True, True,
                                 True, True,  True, True, True]))



DIAGONAL_PROPERTIES = (
    "diagonals_at_node",
    "diagonal_dirs_at_node",
    "diagonal_adjacent_nodes_at_node",
    "d8_adjacent_nodes_at_node",
    "nodes_at_diagonal",
    "nodes_at_d8",
    "d8s_at_node",
    "d8_dirs_at_node",
    # 'd8_status_at_node',
    "length_of_diagonal",
    "length_of_d8",
    "status_at_diagonal",
    "diagonal_status_at_node",
    "active_diagonals",
    "active_diagonal_dirs_at_node",
    "status_at_d8",
    "active_d8",
    "active_d8_dirs_at_node",
)


def pytest_generate_tests(metafunc):
    if "diagonal_property" in metafunc.fixturenames:
        metafunc.parametrize("diagonal_property", DIAGONAL_PROPERTIES)



import pytest
import numpy as np
import math
from numpy.testing import assert_array_equal, assert_almost_equal

from landlab import RasterModelGrid, FieldError, HexModelGrid
from landlab.components import FlowAccumulator, FlowDirectorSteepest
from landlab.utils.flow__distance import calculate_flow__distance

def test_no_flow_recievers():
    """Test that correct error is raised when no flow recievers are
    on the grid."""

    # instantiate a model grid, do not run flow accumulation on it

    mg = RasterModelGrid((30, 70))

    # test that the flow distance utility will fail because of a ValueError

    with pytest.raises(FieldError):
        calculate_flow__distance(mg)


def test_no_upstream_array():
    """Test that correct error is raised when no flow__upstream_node_order."""

    # instantiate a model grid, do not run flow accumulation on it

    mg = RasterModelGrid(30, 70)

    #Add a field called topographic__elevation to mg

    z = mg.add_ones('node','topographic__elevation')

    #Run the FlowDirectorSteepest component

    fd = FlowDirectorSteepest(mg)
    fd.run_one_step()

    # test that the flow distance utility will fail because of a ValueError

    with pytest.raises(FieldError):
        calculate_flow__distance(mg)


def test_flow__distance_regular_grid_d8():
    """Test to demonstrate that flow__distance utility works as expected with
    regular grids"""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), spacing=(1, 1))

    # instantiate an elevation array

    z = np.array([[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0],
                  [0, 0, 0, 0]], dtype='float64')

    # add the elevation field to the grid

    mg.add_field('node', 'topographic__elevation', z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with D8 algorithm

    flow__distance_expected = np.array([[0, 0, 0, 0], [0, 1, 0, 0],
                                       [0, math.sqrt(2), 1, 0],
                                       [0, 1+math.sqrt(2), 2, 0], [0, 0, 0, 0]],
                                       dtype='float64')
    flow__distance_expected = np.reshape(flow__distance_expected,
                                        mg.number_of_node_rows *
                                        mg.number_of_node_columns)

    #setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
                                           left_is_closed=True,
                                           right_is_closed=True,
                                           top_is_closed=True)

    # calculating flow directions with FlowAccumulator component

    fr = FlowAccumulator(mg, flow_director='D8')
    fr.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True,
                                             noclobber=False)
    flow__distance = np.reshape(flow__distance, mg.number_of_node_rows *
                               mg.number_of_node_columns)

    # modifying the flow distance map because boundary and outlet nodes should
    # not have flow__distance value different from 0

    flow__distance[mg.boundary_nodes] = 0
    outlet_id = 6
    flow__distance[outlet_id] = 0


    # test that the flow distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)


def test_flow__distance_regular_grid_d4():
    """Test to demonstrate that flow__distance utility works as expected with
    regular grids"""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), spacing=(1, 1))

    # instantiate an elevation array

    z = np.array([[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0],
                  [0, 0, 0, 0]], dtype='float64')

    # add the elevation field to the grid

    mg.add_field('node', 'topographic__elevation', z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with D4 algorithm

    flow__distance_expected = np.array([[0, 0, 0, 0], [0, 1, 0, 0], [0, 2, 1, 0],
                                        [0, 3, 2, 0], [0, 0, 0, 0]],
                                        dtype='float64')
    flow__distance_expected = np.reshape(flow__distance_expected,
                                         mg.number_of_node_rows *
                                         mg.number_of_node_columns)

    #setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
                                           left_is_closed=True,
                                           right_is_closed=True,
                                           top_is_closed=True)

    # calculating flow directions with FlowAccumulator component

    fr = FlowAccumulator(mg, flow_director='D4')
    fr.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True,
                                          noclobber=False)
    flow__distance = np.reshape(flow__distance,mg.number_of_node_rows *
                                mg.number_of_node_columns)

    # modifying the flow distance map because boundary and outlet nodes
    # should not have flow__distance value different from 0

    flow__distance[mg.boundary_nodes] = 0
    outlet_id = 6
    flow__distance[outlet_id] = 0

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)


def test_flow__distance_irregular_grid_d4():
    """Test to demonstrate that flow__distance utility works as expected with irregular grids"""

    # instantiate a model grid

    dx = 1.0
    hmg = HexModelGrid(5, 3, dx)

    # instantiate and add the elevation field

    hmg.add_field('topographic__elevation', hmg.node_x + np.round(hmg.node_y),
                  at='node')

    # instantiate the expected flow__distance array

    flow__distance_expected = np.array([0., 0., 0.,
                                        0., 0., dx, 0.,
                                        0., dx, dx, 2.*dx, 0.,
                                        0., 2.*dx, 2.*dx, 0.,
                                        0., 0., 0.])

    #setting boundary conditions

    hmg.set_closed_nodes(hmg.boundary_nodes)

    # calculating flow directions with FlowAccumulator component: D4 algorithm

    fr = FlowAccumulator(hmg, flow_director = 'D4')
    fr.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(hmg, add_to_grid=True,
                                              noclobber=False)

    # test that the flow__distance utility works as expected

    assert_almost_equal(flow__distance_expected, flow__distance, decimal=10)


def test_flow__distance_raster_MFD_diagonals_true():
    """Test of flow__distance utility with a raster grid and MFD."""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), spacing=(1, 1))

    # instantiate an elevation array

    z = np.array([[0 ,0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0],
                  [0, 0, 0, 0]], dtype='float64')

    # add the elevation field to the grid

    mg.add_field('node', 'topographic__elevation', z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with MFD algorithm

    flow__distance_expected = np.array([[0, 0, 0, 0], [0, 1, 0, 0],
                                        [0, math.sqrt(2), 1, 0],
                                        [0, 1+math.sqrt(2), 2, 0], [0, 0, 0, 0]],
                                        dtype='float64')
    flow__distance_expected = np.reshape(flow__distance_expected,
                                        mg.number_of_node_rows *
                                        mg.number_of_node_columns)

    #setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
                                           left_is_closed=True,
                                           right_is_closed=True,
                                           top_is_closed=True)

    # calculating flow directions with FlowAccumulator component

    fa = FlowAccumulator(mg, 'topographic__elevation', flow_director='MFD',
                         diagonals=True)
    fa.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True,
                                              noclobber=False)

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)

def test_flow__distance_raster_MFD_diagonals_false():
    """Test of flow__distance utility with a raster grid and MFD."""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), spacing=(1, 1))

    # instantiate an elevation array

    z = np.array([[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0],
                  [0, 0, 0, 0]], dtype='float64')

    # add the elevation field to the grid

    mg.add_field('node', 'topographic__elevation', z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with MFD algorithm

    flow__distance_expected = np.array([[0, 0, 0, 0], [0, 1, 0, 0], [0, 2, 1, 0],
                                        [0, 3, 2, 0], [0, 0, 0, 0]],
                                        dtype='float64')
    flow__distance_expected = np.reshape(flow__distance_expected,
                                        mg.number_of_node_rows *
                                        mg.number_of_node_columns)

    #setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
                                           left_is_closed=True,
                                           right_is_closed=True,
                                           top_is_closed=True)

    # calculating flow directions with FlowAccumulator component

    fa = FlowAccumulator(mg, 'topographic__elevation', flow_director='MFD',
                         diagonals=False)
    fa.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True,
                                              noclobber=False)

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)

def test_flow__distance_raster_D_infinity():
    """Test of flow__distance utility with a raster grid and D infinity."""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), spacing=(1, 1))

    # instantiate an elevation array

    z = mg.x_of_node + 2.0 * mg.y_of_node

    # add the elevation field to the grid

    mg.add_field('node', 'topographic__elevation', z)

    # instantiate the expected flow_length array

    flow__distance_expected = np.array([[0, 0, 0, 0], [0, 0, 0, 0],
                                        [0, 1, 1, math.sqrt(2)],
                                        [0, 2, 2, 1+math.sqrt(2)],
                                        [0, 3, 3, 2+math.sqrt(2)]], dtype='float64')
    flow__distance_expected = np.reshape(flow__distance_expected,
                                        mg.number_of_node_rows *
                                        mg.number_of_node_columns)

    #setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
                                           left_is_closed=True,
                                           right_is_closed=True,
                                           top_is_closed=True)

    # calculating flow directions with FlowAccumulator component

    fa = FlowAccumulator(mg, 'topographic__elevation', flow_director='DINF')
    fa.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True,
                                              noclobber=False)

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)



"""Test HexGraph and DualHexGraph."""
from pytest import approx
from numpy.testing import assert_array_equal, assert_array_almost_equal
from scipy.spatial import Voronoi
import numpy as np

from landlab.graph import HexGraph, DualHexGraph


def test_create_rect():
    """Test creating a hex graph with rectangular layout."""
    graph = HexGraph((3, 2), node_layout='rect')

    assert graph.number_of_nodes == 6
    assert graph.number_of_links == 9
    assert graph.number_of_patches == 4


def test_create_hex():
    """Test creating a hex graph with hex layout."""
    graph = HexGraph((3, 2), node_layout='hex')

    assert graph.number_of_nodes == 7
    assert graph.number_of_links == 12
    assert graph.number_of_patches == 6


def test_create_rect1():
    """Test creating a hex graph."""
    graph = HexGraph((3, 2), node_layout='rect1')

    assert graph.number_of_nodes == 7
    assert graph.number_of_links == 12
    assert graph.number_of_patches == 6


def test_spacing():
    """Test spacing of nodes."""
    graph = HexGraph((20, 31))
    assert_array_almost_equal(graph.length_of_link, 1.)

    graph = HexGraph((31, 20), spacing=2)
    assert_array_almost_equal(graph.length_of_link, 2.)


def test_origin():
    """Test setting the origin."""
    graph = HexGraph((4, 3))

    assert graph.y_of_node[0] == approx(0.)
    assert graph.x_of_node[0] == approx(0.)

    graph = HexGraph((4, 3), origin=(.5, .25))

    assert graph.y_of_node[0] == approx(.5)
    assert graph.x_of_node[0] == approx(.25)


def test_orientation():
    """Test vertical and horizontal orientation."""
    graph = HexGraph((3, 3), orientation='vertical')
    assert_array_almost_equal(graph.y_of_node,
                              [0., 0., .5, 1., 1., 1.5, 2., 2., 2.5])

    graph = HexGraph((3, 3), orientation='horizontal')
    assert_array_almost_equal(graph.x_of_node,
                              [.0, 1., 2., .5, 1.5, 2.5, 0., 1., 2.])


def test_perimeter_nodes_rect():
    graph = HexGraph((3, 4), node_layout='rect')
    assert_array_equal(graph.perimeter_nodes, [3, 7, 11, 10, 9, 8, 4, 0, 1, 2])


def test_perimeter_nodes_hex():
    graph = HexGraph((4, 2), node_layout='hex')
    assert_array_equal(graph.perimeter_nodes, [1, 4, 8, 11, 10, 9, 5, 2, 0])


def test_adjacent_nodes_at_node():
    graph = HexGraph((3, 3), node_layout='hex')
    assert_array_equal(graph.adjacent_nodes_at_node,
                       [[ 1,  4,  3, -1, -1, -1],
                        [ 2,  5,  4,  0, -1, -1],
                        [ 6,  5,  1, -1, -1, -1],
                        [ 4,  7,  0, -1, -1, -1],
                        [ 5,  8,  7,  3,  0,  1],
                        [ 6,  9,  8,  4,  1,  2],
                        [ 9,  5,  2, -1, -1, -1],
                        [ 8,  3,  4, -1, -1, -1],
                        [ 9,  7,  4,  5, -1, -1],
                        [ 8,  5,  6, -1, -1, -1]])



#! /usr/bin/env python
from .grouped import ModelDataFields, GroupSizeError


class ModelDataFieldsMixIn(ModelDataFields):

    """Mix-in that provides un-sized fields.

    Inherit from this class to provide fields that do not have to be given
    a size on instantiation. The size of a particular group will only be set
    after the first time a group is accessed (a field is added to a group or
    an array is created for a group). The size of that group will be the
    size as returned by the `number_of_elements` method.

    This mix-in assumes it is being added to a class that implements a method
    that, given an element name as a string, returns the number of elements
    for that group. A `RasterModelGrid` is an excellent example of this.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((4, 5))
    >>> grid.number_of_elements('node')
    20
    """

    def __init__(self, **kwds):
        super(ModelDataFieldsMixIn, self).__init__(**kwds)

    def empty(self, *args, **kwds):
        """Array, filled with unititialized values, for a given element.

        Returns a numpy array of uninitialized values that is the same length
        as the number of nodes in the grid. Use the *centering* keyword to
        return an array for other elements of the grid. *centering* is a
        string that is one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.

        Parameters
        ----------
        centering : str, optional
            Grid element on which the values are defined.

        Returns
        -------
        ndarray
            A newly-allocated array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 5))
        >>> len(grid.empty())
        20

        LLCATS: FIELDADD
        """
        if len(args) == 0:
            group = kwds.pop('at', kwds.pop('centering', 'node'))
        else:
            group = args[0]
        
        if group=='grid':
            raise ValueError("empty is not supported for at='grid', if you "
                             "want to create a field at the grid, use\n"
                             "grid.at_grid['value_name']=value\n"
                             "instead.")        
        
        n_elements = self.number_of_elements(group)

        if self[group].size is None:
            self[group].size = n_elements
        elif self[group].size != n_elements:
            raise GroupSizeError(group, self[group].size, n_elements)

        return ModelDataFields.empty(self, group, **kwds)

    def ones(self, *args, **kwds):
        """Array, filled with ones, for a given element.

        Returns a numpy array of ones that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.

        Parameters
        ----------
        centering : str, optional
            Grid element on which the values are defined.

        Returns
        -------
        ndarray
            A newly-allocated array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 5))
        >>> grid.zeros(dtype=int) # doctest: +NORMALIZE_WHITESPACE
        array([0, 0, 0, 0, 0,
               0, 0, 0, 0, 0,
               0, 0, 0, 0, 0,
               0, 0, 0, 0, 0])
        >>> grid.zeros('cell') # doctest: +NORMALIZE_WHITESPACE
        array([ 0., 0., 0.,
                0., 0., 0.])

        LLCATS: FIELDADD
        """
        if len(args) == 0:
            group = kwds.pop('at', kwds.pop('centering', 'node'))
        else:
            group = args[0]
        
        if group=='grid':
            raise ValueError("ones is not supported for at='grid', if you "
                             "want to create a field at the grid, use\n"
                             "grid.at_grid['value_name']=value\n"
                             "instead.\nAlternatively, if you want ones"
                             "of the shape stored at_grid, use np.array(1).")
                             
        n_elements = self.number_of_elements(group)

        if self[group].size is None:
            self[group].size = n_elements
        elif self[group].size != n_elements:
            raise GroupSizeError(group, self[group].size, n_elements)

        return ModelDataFields.ones(self, group, **kwds)

    def zeros(self, *args, **kwds):
        """Array, filled with zeros, for a given element.

        Returns a numpy array of zeros that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.

        Parameters
        ----------
        centering : str, optional
            Grid element on which the values are defined.

        Returns
        -------
        ndarray
            A newly-allocated array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 5))
        >>> grid.zeros()
        array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                0.,  0.,  0.,  0.,  0.,  0.,  0.])

        LLCATS: FIELDADD
        """
        if len(args) == 0:
            group = kwds.pop('at', kwds.pop('centering', 'node'))
        else:
            group = args[0]
            
        if group=='grid':
            raise ValueError("zeros is not supported for at='grid', if you "
                             "want to create a field at the grid, use\n"
                             "grid.at_grid['value_name']=value\n"
                             "instead.\nAlternatively, if you want zeros"
                             "of the shape stored at_grid, use np.array(0).")
                             
        n_elements = self.number_of_elements(group)

        if self[group].size is None:
            self[group].size = n_elements
        elif self[group].size != n_elements:
            raise GroupSizeError(group, self[group].size, n_elements)

        return ModelDataFields.zeros(self, group, **kwds)



"""Test the flow router component.

This just tests the functionality of the router in toto - no attempt is made
to test the submodules.
Sinks are tested as part of the lake_mapper.

@author: dejh
"""
# Created on Thurs Nov 12, 2015
import os

from six.moves import range

from pytest import approx
import numpy as np
from numpy.testing import assert_array_equal

import landlab
from landlab import RasterModelGrid, RadialModelGrid, FieldError
from landlab.components.flow_routing import FlowRouter
from landlab import CLOSED_BOUNDARY
from landlab import BAD_INDEX_VALUE as XX


_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_check_fields(dans_grid1):
    """Check to make sure the right fields have been created."""
    fr = FlowRouter(dans_grid1.mg)
    assert_array_equal(dans_grid1.z, dans_grid1.mg.at_node["topographic__elevation"])
    assert_array_equal(np.zeros(25), dans_grid1.mg.at_node["drainage_area"])
    assert_array_equal(np.ones(25), dans_grid1.mg.at_node["water__unit_flux_in"])

    fr = FlowRouter(dans_grid1.mg, dans_grid1.infile)
    assert_array_equal(np.full(25, 2.), dans_grid1.mg.at_node["water__unit_flux_in"])


def test_check_field_input(dans_grid1):
    """Check we can successfully pass water__discharge_in."""
    dans_grid1.mg.add_field(
        "node", "water__unit_flux_in", np.full(25, 3.), units="m**3/s"
    )
    fr = FlowRouter(dans_grid1.mg)
    assert_array_equal(np.full(25, 3.), dans_grid1.mg.at_node["water__unit_flux_in"])
    fr = FlowRouter(dans_grid1.mg, dans_grid1.infile)
    assert_array_equal(np.full(25, 2.), dans_grid1.mg.at_node["water__unit_flux_in"])


def test_accumulate_D8(dans_grid1):
    """Test accumulation works for D8 in a simple scenario."""
    fr = FlowRouter(dans_grid1.mg)
    fr.route_flow()
    assert_array_equal(dans_grid1.A_target, dans_grid1.mg.at_node["drainage_area"])
    assert_array_equal(
        dans_grid1.frcvr_target, dans_grid1.mg.at_node["flow__receiver_node"]
    )
    assert_array_equal(
        dans_grid1.upids_target, dans_grid1.mg.at_node["flow__upstream_node_order"]
    )
    assert_array_equal(
        dans_grid1.links2rcvr_target,
        dans_grid1.mg.at_node["flow__link_to_receiver_node"],
    )
    assert_array_equal(
        dans_grid1.A_target, dans_grid1.mg.at_node["surface_water__discharge"]
    )
    assert_array_equal(
        dans_grid1.steepest_target, dans_grid1.mg.at_node["topographic__steepest_slope"]
    )


def test_variable_Qin(dans_grid1):
    """Test variable Qin field."""
    Qin_local = np.zeros(25, dtype=float)
    Qin_local[13] = 2.
    dans_grid1.mg.add_field("node", "water__unit_flux_in", Qin_local, units="m**3/s")
    fr = FlowRouter(dans_grid1.mg)
    fr.route_flow()
    Qout_local = np.zeros_like(Qin_local)
    Qout_local[10:14] = 200.
    assert_array_equal(Qout_local, dans_grid1.mg.at_node["surface_water__discharge"])
    assert_array_equal(dans_grid1.A_target, dans_grid1.mg.at_node["drainage_area"])
    # note that A DOES NOT CHANGE when messing with Q_in


def test_irreg_topo(dans_grid2):
    """Test D8 routing on a toy irregular topo."""
    fr = FlowRouter(dans_grid2.mg)
    fr.route_flow()
    assert_array_equal(dans_grid2.A_target_D8, dans_grid2.mg.at_node["drainage_area"])
    assert_array_equal(
        dans_grid2.frcvr_target_D8, dans_grid2.mg.at_node["flow__receiver_node"]
    )
    assert_array_equal(
        dans_grid2.upids_target_D8, dans_grid2.mg.at_node["flow__upstream_node_order"]
    )
    assert_array_equal(
        dans_grid2.links2rcvr_target_D8,
        dans_grid2.mg.at_node["flow__link_to_receiver_node"],
    )
    assert dans_grid2.steepest_target_D8 == approx(
        dans_grid2.mg.at_node["topographic__steepest_slope"]
    )


def test_irreg_topo_old(dans_grid2):
    """
    Test D4 routing on a toy irregular topo, old style, where 'method' is
    passed to the run method, not the init.
    """
    fr = FlowRouter(dans_grid2.mg)
    fr.route_flow(method="D4")
    assert_array_equal(dans_grid2.A_target_D4, dans_grid2.mg.at_node["drainage_area"])
    assert_array_equal(
        dans_grid2.frcvr_target_D4, dans_grid2.mg.at_node["flow__receiver_node"]
    )
    assert_array_equal(
        dans_grid2.upids_target_D4, dans_grid2.mg.at_node["flow__upstream_node_order"]
    )
    assert_array_equal(
        dans_grid2.links2rcvr_target_D4,
        dans_grid2.mg.at_node["flow__link_to_receiver_node"],
    )
    assert dans_grid2.steepest_target_D4 == approx(
        dans_grid2.mg.at_node["topographic__steepest_slope"]
    )


def test_irreg_topo_new(dans_grid2):
    """Test D4 routing on a toy irregular topo. 'method' passed to init."""
    fr = FlowRouter(dans_grid2.mg, method="D4")
    fr.route_flow()
    assert_array_equal(dans_grid2.A_target_D4, dans_grid2.mg.at_node["drainage_area"])
    assert_array_equal(
        dans_grid2.frcvr_target_D4, dans_grid2.mg.at_node["flow__receiver_node"]
    )
    assert_array_equal(
        dans_grid2.upids_target_D4, dans_grid2.mg.at_node["flow__upstream_node_order"]
    )
    assert_array_equal(
        dans_grid2.links2rcvr_target_D4,
        dans_grid2.mg.at_node["flow__link_to_receiver_node"],
    )
    assert dans_grid2.steepest_target_D4 == approx(
        dans_grid2.mg.at_node["topographic__steepest_slope"]
    )


def test_internal_closed(internal_closed):
    """Test closed nodes in the core of the grid."""
    fr = FlowRouter(internal_closed.mg)
    fr.route_flow()
    assert internal_closed.A_target == approx(
        internal_closed.mg.at_node["drainage_area"]
    )
    assert_array_equal(
        internal_closed.frcvr_target, internal_closed.mg.at_node["flow__receiver_node"]
    )
    assert_array_equal(
        internal_closed.links2rcvr_target,
        internal_closed.mg.at_node["flow__link_to_receiver_node"],
    )
    assert internal_closed.A_target == approx(
        internal_closed.mg.at_node["surface_water__discharge"]
    )
    assert internal_closed.steepest_target == approx(
        internal_closed.mg.at_node["topographic__steepest_slope"]
    )


def test_voronoi():
    """Test routing on a (radial) voronoi."""
    vmg = RadialModelGrid(2, dr=2.)
    z = np.full(20, 10., dtype=float)
    # vmg.status_at_node[8:] = CLOSED_BOUNDARY
    all_bounds_but_one = np.array((0, 1, 2, 3, 4, 7, 11, 15, 16, 17, 18, 19))
    vmg.status_at_node[all_bounds_but_one] = CLOSED_BOUNDARY
    # z[7] = 0.  # outlet
    z[12] = 0.  # outlet
    # inner_elevs = (3., 1., 4., 5., 6., 7., 8.)
    inner_elevs = (8., 7., 3., 1., 6., 4., 5.)
    # z[:7] = np.array(inner_elevs)
    z[vmg.core_nodes] = np.array(inner_elevs)
    vmg.add_field("node", "topographic__elevation", z, units="-")
    fr = FlowRouter(vmg)

    #    nodes_contributing = [np.array([0, 3, 4, 5]),
    #                          np.array([0, 1, 2, 3, 4, 5, 6]),
    #                          np.array([2, ]),
    #                          np.array([3, ]),
    #                          np.array([4, ]),
    #                          np.array([5, ]),
    #                          np.array([6, ])]

    # The follow list contains arrays with the IDs of cells contributing flow
    # to nodes 5, 6, 8, 9, 10, 13, and 14, respectively (which correspond to
    # cells 0-6)
    cells_contributing = [
        np.array([0]),
        np.array([1]),
        np.array([1, 2, 4, 6]),
        np.array([0, 1, 2, 3, 4, 5, 6]),
        np.array([4]),
        np.array([5]),
        np.array([6]),
    ]

    A_target_core = np.zeros(vmg.number_of_core_nodes)
    for i in range(7):
        A_target_core[i] = vmg.area_of_cell[cells_contributing[i]].sum()
    A_target_outlet = vmg.area_of_cell.sum()

    fr.route_flow()
    assert vmg.at_node["drainage_area"][vmg.core_nodes] == approx(A_target_core)
    assert vmg.at_node["drainage_area"][12] == approx(A_target_outlet)


def test_voronoi_closedinternal():
    """Test routing on a (radial) voronoi, but with a closed interior node."""
    vmg = RadialModelGrid(2, dr=2.)
    z = np.full(20, 10., dtype=float)
    all_bounds_but_one = np.array((0, 1, 2, 3, 4, 7, 11, 15, 16, 17, 18, 19))
    vmg.status_at_node[all_bounds_but_one] = CLOSED_BOUNDARY
    vmg.status_at_node[8] = CLOSED_BOUNDARY  # new internal closed
    z[12] = 0.  # outlet
    inner_elevs = (8., 7., 1., 6., 4., 5.)
    z[vmg.core_nodes] = np.array(inner_elevs)
    vmg.add_field("node", "topographic__elevation", z, units="-")
    fr = FlowRouter(vmg)

    cells_contributing = [
        np.array([0]),
        np.array([1]),
        np.array([0, 1, 3, 4, 5, 6]),
        np.array([1, 4]),
        np.array([1, 4, 5, 6]),
        np.array([1, 4, 6]),
    ]

    A_target_internal = np.zeros(vmg.number_of_core_nodes, dtype=float)
    for i in range(6):
        A_target_internal[i] = vmg.area_of_cell[cells_contributing[i]].sum()
    A_target_outlet = vmg.area_of_cell[vmg.cell_at_node[vmg.core_nodes]].sum()
    fr.route_flow()

    assert vmg.at_node["drainage_area"][vmg.core_nodes] == approx(A_target_internal)
    assert vmg.at_node["drainage_area"][12] == approx(A_target_outlet)



#! /usr/bin/env python
"""Unit tests for landlab.io.netcdf module."""
import pytest
import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.io.netcdf import write_netcdf, NotRasterGridError, WITH_NETCDF4
from landlab.io.netcdf.read import _get_raster_spacing

try:
    import netCDF4 as nc
except ImportError:
    pass


def test_netcdf_write_int64_field_netcdf4(tmpdir):
    """Test write_netcdf with a grid that has an int64 field."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation',
                    np.arange(12, dtype=np.int64))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, format='NETCDF4')

        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation']:
            assert name in root.variables
            assert_array_equal(root.variables[name][:].flatten(),
                               field.at_node[name])
            assert root.variables[name][:].dtype == 'int64'

        root.close()


def test_netcdf_write_uint8_field_netcdf4(tmpdir):
    """Test write_netcdf with a grid that has an uint8 field."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation',
                    np.arange(12, dtype=np.uint8))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, format='NETCDF4')

        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation']:
            assert name in root.variables
            assert_array_equal(root.variables[name][:].flatten(),
                               field.at_node[name])
            assert root.variables[name][:].dtype == 'uint8'

        root.close()


def test_netcdf_write_as_netcdf3_64bit(tmpdir):
    """Test write_netcdf with output format 64-bit netcdf3."""
    from scipy.io import netcdf

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', 2. * np.arange(12.))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, format='NETCDF3_64BIT')

        f = netcdf.netcdf_file('test.nc', 'r')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert name in f.variables
            assert_array_equal(f.variables[name][:].flatten(), field.at_node[name])

        f.close()


def test_netcdf_write_as_netcdf3_classic(tmpdir):
    """Test write_netcdf with output format classic netcdf3."""
    from scipy.io import netcdf

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', 2. * np.arange(12.))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, format='NETCDF3_CLASSIC')

        f = netcdf.netcdf_file('test.nc', 'r')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert name in f.variables
            assert_array_equal(f.variables[name][:].flatten(), field.at_node[name])

        f.close()


@pytest.mark.skipif(not WITH_NETCDF4, reason='netCDF4 package not installed')
def test_netcdf_write(tmpdir):
    """Test generic write_netcdf."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        assert set(root.dimensions) == set(['ni', 'nj', 'nt'])
        assert len(root.dimensions['ni']) == 3
        assert len(root.dimensions['nj']) == 4
        assert len(root.dimensions['nt']) == 1
        assert root.dimensions['nt'].isunlimited()

        assert set(root.variables) == set(['x', 'y', 'topographic__elevation'])

        assert_array_equal(root.variables['x'][:].flatten(),
                           np.array([0., 1., 2., 0., 1., 2., 0., 1., 2.,
                                     0., 1., 2., ]))
        assert_array_equal(root.variables['y'][:].flatten(),
                           np.array([0., 0., 0., 1., 1., 1., 2., 2., 2.,
                                     3., 3., 3., ]))
        assert_array_equal(root.variables['topographic__elevation'][:].flatten(),
                           field.at_node['topographic__elevation'])

        root.close()


@pytest.mark.skipif(not WITH_NETCDF4, reason='netCDF4 package not installed')
def test_netcdf_write_as_netcdf4_classic(tmpdir):
    """Test write_netcdf to netcdf4 classic format."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, format='NETCDF4_CLASSIC')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4_CLASSIC')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert name in root.variables
            assert_array_equal(root.variables[name][:].flatten(),
                               field.at_node[name])

        root.close()


@pytest.mark.skipif(not WITH_NETCDF4, reason='netCDF4 package not installed')
def test_netcdf_write_names_keyword_as_list(tmpdir):
    """Test write_netcdf using a list for the *names* keyword."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, names=['topographic__elevation'],
                     format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        assert 'topographic__elevation' in root.variables
        assert 'uplift_rate' not in root.variables
        assert_array_equal(root.variables['topographic__elevation'][:].flatten(),
                           field.at_node['topographic__elevation'])

        root.close()


@pytest.mark.skipif(not WITH_NETCDF4, reason='netCDF4 package not installed')
def test_netcdf_write_names_keyword_as_str(tmpdir):
    """Test write_netcdf using a ``str`` for the *names* keyword."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, names='uplift_rate', format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        assert 'topographic__elevation' not in root.variables
        assert 'uplift_rate' in root.variables
        assert_array_equal(root.variables['uplift_rate'][:].flatten(),
                           field.at_node['uplift_rate'])

        root.close()


@pytest.mark.skipif(not WITH_NETCDF4, reason='netCDF4 package not installed')
def test_netcdf_write_names_keyword_as_none(tmpdir):
    """Test write_netcdf using ``None`` for the *names* keyword."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with tmpdir.as_cwd():
        write_netcdf('test.nc', field, names=None, format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert name in root.variables
            assert_array_equal(root.variables[name][:].flatten(),
                               field.at_node[name])

        root.close()


def test_2d_unit_spacing():
    """Test write_netcdf with a 2D grid with unit spacing."""
    (x, y) = np.meshgrid(np.arange(5.), np.arange(4.))

    spacing = _get_raster_spacing((y, x))
    assert spacing == 1.


def test_2d_non_unit_spacing():
    """Test _get_raster_spacing with a 2D grid with non-unit spacing."""
    (x, y) = np.meshgrid(np.arange(5.) * 2, np.arange(4.) * 2)

    spacing = _get_raster_spacing((y, x))
    assert spacing == 2.


def test_2d_uneven_spacing_axis_0():
    """Test _get_raster_spacing with a 2D grid with uneven spacing in y."""
    (x, y) = np.meshgrid(np.logspace(0., 2., num=5), np.arange(4.))

    with pytest.raises(NotRasterGridError):
        _get_raster_spacing((y, x))


def test_2d_uneven_spacing_axis_1():
    """Test _get_raster_spacing with a 2D grid with uneven spacing in x."""
    (x, y) = np.meshgrid(np.arange(4.), np.logspace(0., 2., num=5))

    with pytest.raises(NotRasterGridError):
        _get_raster_spacing((y, x))


def test_2d_switched_coords():
    """Test _get_raster_spacing with a 2D grid when the spacing is switched."""
    (x, y) = np.meshgrid(np.arange(5.), np.arange(4.))

    spacing = _get_raster_spacing((x, y))
    assert spacing == 0.


def test_1d_unit_spacing():
    """Test _get_raster_spacing with a 1D grid with unit spacing."""
    spacing = _get_raster_spacing((np.arange(5.), ))
    assert spacing == 1.


def test_1d_non_unit_spacing():
    """Test _get_raster_spacing with a 1D grid with non-unit spacing."""
    spacing = _get_raster_spacing((np.arange(5.) * 2, ))
    assert spacing == 2.


def test_1d_uneven_spacing():
    """Test _get_raster_spacing with a 1D grid with uneven spacing in y."""
    with pytest.raises(NotRasterGridError):
        _get_raster_spacing((np.logspace(0., 2., num=5), ))


@pytest.mark.skipif(not WITH_NETCDF4, reason='netCDF4 package not installed')
def test_netcdf_write_at_cells(tmpdir):
    """Test write_netcdf using with cell fields"""
    field = RasterModelGrid((4, 3))
    field.add_field('cell', 'topographic__elevation',
                    np.arange(field.number_of_cells))
    field.add_field('cell', 'uplift_rate', np.arange(field.number_of_cells))

    with tmpdir.as_cwd():
        write_netcdf('test-cells.nc', field, format='NETCDF4')
        root = nc.Dataset('test-cells.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert name in root.variables
            assert_array_equal(root.variables[name][:].flatten(),
                               field.at_cell[name])

        assert set(root.dimensions) == set(['nv', 'ni', 'nj', 'nt'])
        assert len(root.dimensions['nv']) == 4
        assert len(root.dimensions['ni']) == 1
        assert len(root.dimensions['nj']) == 2
        assert len(root.dimensions['nt']) == 1
        assert root.dimensions['nt'].isunlimited()

        assert set(root.variables) == set(['x_bnds', 'y_bnds', 'topographic__elevation', 'uplift_rate'])
        root.close()



import numpy as np


def get_area_of_patch(graph, out=None):
    from .ext.of_patch import calc_area_at_patch

    if out is None:
        out = np.empty(graph.number_of_patches, dtype=float)

    calc_area_at_patch(graph.nodes_at_patch,
                       np.ascontiguousarray(graph.x_of_node),
                       np.ascontiguousarray(graph.y_of_node),
                       out)

    return out


def get_centroid_of_patch(graph, out=None):
    from .ext.of_patch import calc_centroid_at_patch

    if out is None:
        out = np.empty((graph.number_of_patches, 2), dtype=float)

    calc_centroid_at_patch(graph.nodes_at_patch,
                           np.ascontiguousarray(graph.x_of_node),
                           np.ascontiguousarray(graph.y_of_node),
                           out)

    return out



import numpy as np

from ..voronoi.voronoi import VoronoiGraph


def number_of_nodes(shape):
    return np.sum(np.arange(1, shape[0] + 1)) * shape[1] + 1


def create_xy_of_node(shape, spacing=1., origin=(0., 0.)):
    n_shells, n_points = shape
    n_nodes = number_of_nodes(shape)

    x = np.empty((n_nodes, ), dtype=float)
    y = np.empty((n_nodes, ), dtype=float)

    x[0] = y[0] = 0.
    offset = 1
    for shell in range(1, n_shells + 1):
        rho = spacing * shell
        d_theta = np.pi * 2 / (shell * shape[1])
        theta = np.arange(shell * shape[1]) * d_theta

        y[offset: offset + len(theta)] = rho * np.sin(theta)
        x[offset: offset + len(theta)] = rho * np.cos(theta)

        offset += len(theta)

    x = np.round(x, decimals=6)
    y = np.round(y, decimals=6)

    y += origin[0]
    x += origin[1]

    return (x, y)


class RadialGraph(VoronoiGraph):

    """Graph of a series of points on concentric circles.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import RadialGraph
    >>> graph = RadialGraph((1, 4))
    >>> graph.number_of_nodes == 5
    True
    >>> graph.y_of_node
    array([-1.,  0.,  0.,  0.,  1.])
    >>> graph.x_of_node
    array([ 0., -1.,  0.,  1.,  0.])
    """

    def __init__(self, shape, spacing=1., origin=(0., 0.)):
        """Create a structured grid of triangles arranged radially.

        Parameters
        ----------
        shape : tuple of int
            Shape of the graph as number of rings and number of points
            in the first ring.
        spacing : float, optional
            Spacing between rings.
        origin : tuple of float, optional
            Coordinates of the center of the grid.
        """
        try:
            spacing = float(spacing)
        except TypeError:
            raise TypeError('spacing must be a float')

        x_of_node, y_of_node = create_xy_of_node(shape, spacing=spacing,
                                                 origin=origin)

        super(RadialGraph, self).__init__(
            (y_of_node, x_of_node), xy_sort=True, rot_sort=True)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab import FIXED_VALUE_BOUNDARY as FV
from landlab import CLOSED_BOUNDARY as CB


def test_get_status():
    """Test getting the node status array."""
    grid = RasterModelGrid((4, 5))
    assert_array_equal(grid.status_at_node,
                       [FV, FV, FV, FV, FV,
                        FV,  0,  0,  0, FV,
                        FV,  0,  0,  0, FV,
                        FV, FV, FV, FV, FV])


def test_set_status_with_scalar():
    """Test setting status with a scalar."""
    grid = RasterModelGrid((4, 5))
    grid.status_at_node[6] = 2

    assert_array_equal(grid.status_at_node,
                       [FV, FV, FV, FV, FV,
                        FV,  2,  0,  0, FV,
                        FV,  0,  0,  0, FV,
                        FV, FV, FV, FV, FV])


def test_set_status_with_slice():
    """Test setting status with an array slice."""
    grid = RasterModelGrid((4, 5))
    grid.status_at_node[6: 9] = 2

    assert_array_equal(grid.status_at_node,
                       [FV, FV, FV, FV, FV,
                        FV,  2,  2,  2, FV,
                        FV,  0,  0,  0, FV,
                        FV, FV, FV, FV, FV])


def test_set_status_with_array_bool():
    """Test setting node status with boolean array."""
    grid = RasterModelGrid((4, 5))
    inds = np.full((20, ), False, dtype=bool)
    inds[6] = True
    inds[7] = True
    inds[13] = True

    grid.status_at_node[inds] = 2
    assert_array_equal(grid.status_at_node,
                       [FV, FV, FV, FV, FV,
                        FV,  2,  2,  0, FV,
                        FV,  0,  0,  2, FV,
                        FV, FV, FV, FV, FV])


def test_set_with_itemset():
    grid = RasterModelGrid((4, 5))
    grid.status_at_node.itemset(7, 2)

    assert_array_equal(grid.status_at_node,
                       [FV, FV, FV, FV, FV,
                        FV,  0,  2,  0, FV,
                        FV,  0,  0,  0, FV,
                        FV, FV, FV, FV, FV])


def test_set_status_with_array():
    """Test that active links are reset after changing the node status."""
    grid = RasterModelGrid((4, 5))

    assert_array_equal(grid.active_links,
                      [ 5,  6,  7,  9, 10, 11, 12, 14, 15,
                       16, 18, 19, 20, 21, 23, 24, 25])

    grid.status_at_node[: 5] = CB
    assert_array_equal(grid.status_at_node,
                       [CB, CB, CB, CB, CB,
                        FV,  0,  0,  0, FV,
                        FV,  0,  0,  0, FV,
                        FV, FV, FV, FV, FV])

    assert_array_equal(grid.active_links,
                      [ 9, 10, 11, 12, 14, 15, 16, 18, 19, 20, 21, 23, 24, 25])



#! /usr/env/python

"""
flow_director_dinf.py: provides the component FlowDirectorDINF.

Directs flow on raster grids only using the Dinfinity algorithm of
Tarboton 1997.
"""

from landlab.components.flow_director.flow_director_to_many import(
_FlowDirectorToMany)
from landlab.components.flow_director import flow_direction_dinf
from landlab import VoronoiDelaunayGrid
from landlab import (FIXED_VALUE_BOUNDARY, FIXED_GRADIENT_BOUNDARY,
                     BAD_INDEX_VALUE)
import numpy


class FlowDirectorDINF(_FlowDirectorToMany):

    """
    Flow direction on a raster grid by the D infinity method.

    Directs flow by the D infinity method (Tarboton, 1997). Each node is
    assigned two flow directions, toward the two neighboring nodes that are on
    the steepest subtriangle. Partitioning of flow is done based on the aspect
    of the subtriangle.

    Specifically, it stores as ModelGrid fields:

    -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
       there is no receiver: *'flow__receiver_nodes'*. This array is 2D, and is
       of dimension (number of nodes x max number of receivers).
    -  Node array of flow proportions: *'flow__receiver_proportions'*. This
       array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of links carrying flow:  *'flow__links_to_receiver_nodes'*.
       This array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of the steepest downhill receiver. *'flow__receiver_nodes'*
    -  Node array of steepest downhill slope from each receiver:
       *'topographic__steepest_slope'*
    -  Node array containing ID of steepest link that leads from each node to a
       receiver, or BAD_INDEX_VALUE if no link:
       *'flow__link_to_receiver_node'*
    -  Boolean node array of all local lows: *'flow__sink_flag'*

    The primary method of this class is :func:`run_one_step`.

    Examples
    --------

    This method works for both raster and irregular grids. First we will look
    at a raster example, and then an irregular example.

    >>> import numpy as numpy
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowDirectorDINF
    >>> mg = RasterModelGrid((4,4), spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field('topographic__elevation',
    ...                  mg.node_x**2 + mg.node_y**2,
    ...                  at = 'node')

    The DINF flow director can be uses for raster grids only.

    >>> fd = FlowDirectorDINF(mg, 'topographic__elevation')
    >>> fd.surface_values.reshape(mg.shape)
    array([[  0.,   1.,   4.,   9.],
           [  1.,   2.,   5.,  10.],
           [  4.,   5.,   8.,  13.],
           [  9.,  10.,  13.,  18.]])
    >>> fd.run_one_step()

    Unlike flow directors that only direct flow to one node or to all
    downstream nodes, FlowDirectorDINF directs flow two nodes only. It stores
    the receiver information is a (number of nodes x 2) shape field at nodes.

    >>> mg.at_node['flow__receiver_nodes']
    array([[ 0, -1],
           [ 1, -1],
           [ 2, -1],
           [ 3, -1],
           [ 0,  1],
           [ 1,  0],
           [ 5,  1],
           [ 6,  2],
           [ 8, -1],
           [ 5, -1],
           [ 6,  5],
           [10,  6],
           [12, -1],
           [ 9, -1],
           [10,  9],
           [-1, 10]])

    It also stores the proportions of flow going to each receiver and the link
    on which the flow moves in at node arrays.

    >>> mg.at_node['flow__receiver_proportions'] # doctest: +NORMALIZE_WHITESPACE
    array([[ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.        ,  1.        ],
           [ 0.59033447,  0.40966553],
           [ 0.74866817,  0.25133183],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.        ,  1.        ],
           [ 0.31191652,  0.68808348],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.31191652,  0.68808348],
           [ 0.        ,  1.        ]])
    >>> mg.at_node['flow__links_to_receiver_nodes']
    array([[-1, -1],
           [-1, -1],
           [-1, -1],
           [-1, -1],
           [ 3, 25],
           [ 4, 24],
           [ 8, 26],
           [ 9, 28],
           [-1, -1],
           [11, 30],
           [12, 32],
           [16, 34],
           [-1, -1],
           [18, 36],
           [19, 38],
           [20, 40]])

    Like flow directors that only direct flow to one downstream node,
    FlowDirectorDINF identifies and stores the steepest slope leading downhill
    from each node, the link carrying that flow, and the receiver receiving
    flow on that link.

    >>> mg.at_node['topographic__steepest_slope'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0.        ,  0.        ,  0.        ,  0.        ,  1.        ,
            1.41421356,  3.        ,  5.        ,  0.        ,  3.        ,
            4.24264069,  5.65685425,  0.        ,  5.        ,  5.65685425,
            7.07106781])
    >>> mg.at_node['flow__link_to_receiver_node']
    array([-1,  0,  1,  2,  3, 24,  8,  9, -1, 11, 32, 34, -1, 18, 38, 40])
    >>> mg.at_node['flow__receiver_node']
    array([ 0,  0,  1,  2,  0,  0,  5,  6,  8,  5,  5,  6, 12,  9,  9, 10])

    Finally, FlowDirectorDINF identifies sinks, or local lows.

    >>> mg.at_node['flow__sink_flag']
    array([1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0], dtype=int8)

    The flow directors also have the ability to return the flow receiver nodes
    through a function called direct_flow()

    >>> fd = FlowDirectorDINF(mg, 'topographic__elevation')
    >>> fd.run_one_step()
    >>> receivers, proportions = fd.direct_flow()
    >>> receivers
    array([[ 0, -1],
           [ 1, -1],
           [ 2, -1],
           [ 3, -1],
           [ 0,  1],
           [ 1,  0],
           [ 5,  1],
           [ 6,  2],
           [ 8, -1],
           [ 5, -1],
           [ 6,  5],
           [10,  6],
           [12, -1],
           [ 9, -1],
           [10,  9],
           [-1, 10]])
    >>> proportions # doctest: +NORMALIZE_WHITESPACE
    array([[ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.        ,  1.        ],
           [ 0.59033447,  0.40966553],
           [ 0.74866817,  0.25133183],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.        ,  1.        ],
           [ 0.31191652,  0.68808348],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.31191652,  0.68808348],
           [ 0.        ,  1.        ]])

    For each donor node (represented by each row) the proportions should sum to
    one.

    >>> proportions.sum(axis=1)
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
            1.,  1.,  1.])

    """

    _name = 'FlowDirectorDINF'

    def __init__(self, grid, surface='topographic__elevation'):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        surface : field name at node or array of length node, optional
            The surface to direct flow across, default is field at node:
            topographic__self.surface_valuesation.
        partition_method: string, optional
            Method for partitioning flow. Options include 'slope' (default) and
            'square_root_of_slope'.
        """

        self.method = 'DINF'
        super(FlowDirectorDINF, self).__init__(grid, surface)
        self._is_Voroni = isinstance(self._grid, VoronoiDelaunayGrid)
        if self._is_Voroni:
            raise NotImplementedError('FlowDirectorDINF is not implemented'
                                      ' for irregular grids.')
        self.max_receivers = 2
        self.updated_boundary_conditions()

        # set the number of recievers, proportions, and receiver links with the
        # right size.
        self.receivers = grid.add_field('flow__receiver_nodes',
                                        BAD_INDEX_VALUE*numpy.ones((self._grid.number_of_nodes,
                                                                    self.max_receivers),
                                                                   dtype=int),
                                        at='node',
                                        dtype=int,
                                        noclobber=False)

        self.receiver_links = grid.add_field('flow__links_to_receiver_nodes',
                                             BAD_INDEX_VALUE*numpy.ones((self._grid.number_of_nodes,
                                                                         self.max_receivers),
                                                                        dtype=int),
                                        at='node',
                                        dtype=int,
                                        noclobber=False)

        self.proportions = grid.add_field('flow__receiver_proportions',
                                        BAD_INDEX_VALUE*numpy.ones((self._grid.number_of_nodes,
                                                                    self.max_receivers),
                                                                   dtype=float),
                                        at='node',
                                        dtype=int,
                                        noclobber=False)


    def updated_boundary_conditions(self):
        """
        Method to update FlowDirectorDINF when boundary conditions change.

        Call this if boundary conditions on the grid are updated after the
        component is instantiated.
        """
        self._active_links = self.grid.active_links
        self._activelink_tail = self.grid.node_at_link_tail[self.grid.active_links]
        self._activelink_head = self.grid.node_at_link_head[self.grid.active_links]

    def run_one_step(self):
        """
        Find flow directions and save to the model grid.

        run_one_step() checks for updated boundary conditions, calculates
        slopes on links, finds basself.surface_valuesel nodes based on the status at node,
        calculates flow directions, and saves results to the grid.

        An alternative to direct_flow() is direct_flow() which does the same
        things but also returns the receiver nodes not return values.
        """
        self.direct_flow()

    def direct_flow(self):
        """
        Find flow directions, save to the model grid, and return receivers.

        direct_flow() checks for updated boundary conditions, calculates
        slopes on links, finds basself.surface_valuesel nodes based on the status at node,
        calculates flow directions, saves results to the grid, and returns a
        at-node array  of receiver nodes. This array is stored in the grid at:
        grid['node']['flow__receiver_nodes']

        An alternative to direct_flow() is run_one_step() which does the same
        things but also returns a at-node array  of receiver nodes. This array
        is stored in the grid at:
        grid['node']['flow__receiver_nodes']
        """
        # step 0. Check and update BCs
        if self._bc_set_code != self.grid.bc_set_code:
            self.updated_boundary_conditions()
            self._bc_set_code = self.grid.bc_set_code

        # Step 1. Find and save base level nodes.
        (baselevel_nodes, ) = numpy.where(
            numpy.logical_or(self._grid.status_at_node == FIXED_VALUE_BOUNDARY,
                             self._grid.status_at_node == FIXED_GRADIENT_BOUNDARY))

        # Calculate flow directions
        (self.receivers, self.proportions, steepest_slope,
        steepest_receiver, sink,
        receiver_links, steepest_link)= \
        flow_direction_dinf.flow_directions_dinf(self._grid,
                                                 self.surface_values,
                                                 baselevel_nodes=baselevel_nodes)

        # Save the four ouputs of this component.
        self._grid['node']['flow__receiver_nodes'][:] = self.receivers
        self._grid['node']['flow__receiver_node'][:] = steepest_receiver
        self._grid['node']['flow__receiver_proportions'][:] = self.proportions
        self._grid['node']['topographic__steepest_slope'][:] = steepest_slope
        self._grid['node']['flow__link_to_receiver_node'][:] = steepest_link
        self._grid['node']['flow__links_to_receiver_nodes'][:] = receiver_links
        self._grid['node']['flow__sink_flag'][:] = numpy.zeros_like(steepest_link,
                                                                    dtype=bool)
        self._grid['node']['flow__sink_flag'][sink] = True

        return (self.receivers, self.proportions)

    @property
    def nodes_receiving_flow(self):
        """Return the node ids of the nodes receiving flow."""
        return self._grid['node']['flow__receiver_nodes']

    @property
    def proportions_of_flow(self):
        """Return the proportion of flow going to receivers."""
        return self._grid['node']['flow__receiver_proportions']

if __name__ == '__main__':
    import doctest
    doctest.testmod()
